# coding: utf-8

"""
    PKS

    PKS API  # noqa: E501

    OpenAPI spec version: 1.1.0
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""


import pprint
import re  # noqa: F401

import six


class UpdateClusterParameters(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'kubernetes_worker_instances': 'int'
    }

    attribute_map = {
        'kubernetes_worker_instances': 'kubernetes_worker_instances'
    }

    def __init__(self, kubernetes_worker_instances=None):  # noqa: E501
        """UpdateClusterParameters - a model defined in Swagger"""  # noqa: E501

        self._kubernetes_worker_instances = None
        self.discriminator = None

        self.kubernetes_worker_instances = kubernetes_worker_instances

    @property
    def kubernetes_worker_instances(self):
        """Gets the kubernetes_worker_instances of this UpdateClusterParameters.  # noqa: E501


        :return: The kubernetes_worker_instances of this UpdateClusterParameters.  # noqa: E501
        :rtype: int
        """
        return self._kubernetes_worker_instances

    @kubernetes_worker_instances.setter
    def kubernetes_worker_instances(self, kubernetes_worker_instances):
        """Sets the kubernetes_worker_instances of this UpdateClusterParameters.


        :param kubernetes_worker_instances: The kubernetes_worker_instances of this UpdateClusterParameters.  # noqa: E501
        :type: int
        """
        if kubernetes_worker_instances is None:
            raise ValueError("Invalid value for `kubernetes_worker_instances`, must not be `None`")  # noqa: E501
        if kubernetes_worker_instances is not None and kubernetes_worker_instances < 1:  # noqa: E501
            raise ValueError("Invalid value for `kubernetes_worker_instances`, must be a value greater than or equal to `1`")  # noqa: E501

        self._kubernetes_worker_instances = kubernetes_worker_instances

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, UpdateClusterParameters):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other
